#include <stdio.h>
#include <string.h>
#include "mpi.h"


main( int argc, char **argv )
{
#  if defined( DEBUG ) || defined( DEBUG_stdout )
#     include "debug.c.h"
#  endif

#  if defined( TABLE )
#     include "table.c.h"
#  endif

#  if defined( GPLOT )
#     include "gplot.c.h"
#  endif

#  if defined( DEMO )
#     include "demo.c.h"
#  endif

   const int     iter_Max = 100;

   int           Ndim;               /*  Dimensionally of the grid           */

   const int     MD_type = MPI_DOUBLE_PRECISION;
   double       *Task_xGrid;         /* Data of Local Task Grid              */
   int          *Task_Grid_Modify;   /* Boolean bit indicates if Modifiable  */

   MPI_Comm      RGX_Comm;           /* Communicator of all sub-grid         */
   int           my_tid;             /* my task ID                           */
   int           N_task;             /* No. of process/tasks in Comm         */
   int          *Glb_Grid_sz;        /* Global grid sizes: Glb_Grid_sz(Ndim) */
   int          *task_Ngbrs;         /* the tid of the nearest neighbor task */
                                     /* task_Ngbrs( 2, Ndim )                */
   int          *task_Grid_endpts;   /* local sub-grid's end points          */
                                     /* task_Grid_endpts( Ndim, 2 )          */
   int          *N_task_1dims;       /* No of task per dimension             */
                                     /* N_task_1dims( Ndim )                 */
   int          *stride_width;       /* width of the stride of ghost point   */
                                     
   int          *task_bndry_strides; /* local task/grid's boundary strides   */
                                     /* task_bndry_strides( Ndim )           */
   int          *task_bndry_lgths;   /* local task/grid's boundary lengths   */
                                     /* task_bndry_lgths( Ndim )             */

   int           N_charges;          /* No. of Charges                       */
   int          *charges_pos;        /* Position ( coordinates ) of Charges  */
                                     /* charges( Ndim, N_charges )           */
   double       *charges_pot;        /* Potential of the charges             */
   double        fbackground;        /* background potential set on boundary */

/* Dummy Variables */
   int          hyper_vol;
   int          iter_idx;
   int          rc, ierr;
   int          idir, ss;



/* Init the MPI */
   MPI_Init( &argc, &argv );

/* Input from user to determine the following :                        */
/* the Dimensionality of the Grid, Ndim.                               */

   ierr = MPI_Comm_rank( MPI_COMM_WORLD, &my_tid );
   if ( my_tid == 0 )
   {
      fprintf( stdout, "Enter the Dimensionality of the Grid, Ndim : " );
      fscanf(  stdin , "%d", &Ndim );
      fprintf( stdout, "\n" );
   }
   ierr = MPI_Bcast( &Ndim, 1, MPI_INT, 0, MPI_COMM_WORLD );

/* Allocate memory for variables depending on Ndim */

   Glb_Grid_sz         = ( int* ) calloc( Ndim, sizeof( int ) );
   task_Ngbrs          = ( int* ) calloc( 2 * Ndim, sizeof( int ) );
   task_Grid_endpts    = ( int* ) calloc( Ndim * 2, sizeof( int ) );
   N_task_1dims        = ( int* ) calloc( Ndim, sizeof( int ) );
   stride_width        = ( int* ) calloc( Ndim, sizeof( int ) );
   task_bndry_strides  = ( int* ) calloc( Ndim, sizeof( int ) );
   task_bndry_lgths    = ( int* ) calloc( Ndim, sizeof( int ) );

/* Create Cartesian Regular Grids : RGX_Comm */

   ierr = RGX_Create_Task_GridS( MPI_COMM_WORLD, Ndim, &N_task,
                                 N_task_1dims, &RGX_Comm );

/* Input from user to determine the following :                        */
/* the size of the global grid, Glb_Grid_sz[],                         */
/* and Set the widths of the strides, stride_width[].                  */

   if ( my_tid == 0 )
   {
      for ( idir = 0; idir < Ndim; idir++ )
      {
          fprintf( stdout, "Enter the Glb_Grid_sz[ %d ] = ", idir );
          fscanf(  stdin , "%d", &Glb_Grid_sz[ idir ] );
          /* fprintf( stdout, "\n" );  */
          stride_width[ idir ] = 1;
      }
      fprintf( stdout, "\n" );
   }
   ierr = RGX_Set_Glb_Grid_Ppty( RGX_Comm, Ndim, Glb_Grid_sz, stride_width );
   
/* Input from user to determine the following :                        */
/* Total No. of Point Charges, N_charges.                              */

   if ( my_tid == 0 )
   {
      fprintf( stdout, "Enter Total No. of Point Charges, N_charges : " );
      fscanf(  stdin , "%d", &N_charges );
      fprintf( stdout, "\n" );
   }
   ierr = MPI_Bcast( &N_charges, 1, MPI_INT, 0, RGX_Comm );
  
/* Allocate memory for variables depending on N_charges */

   charges_pos         = ( int* ) calloc( Ndim * N_charges, sizeof( int ) );
   charges_pot         = ( double* ) calloc( N_charges, sizeof( double ) );

/* Input from user to determine the following :                        */
/* The charges' coordinates, charges_pos[], and the charges' potential */

   if ( my_tid == 0 )
   {
      for ( ss = 0; ss < N_charges; ss++ )
      {
         fprintf( stdout,
                  "Enter Charge {%d}'s Coordinate, separated by blank : ", ss );
         for ( idir = 0; idir < Ndim; idir++ )
            fscanf( stdin, "%d", &charges_pos[ ss * Ndim + idir ] );
         /* fprintf( stdout, "\n" );  */
         fprintf( stdout, "Enter the Potential of the Charge {%d} : ", ss );
         fscanf( stdin, "%lf", &charges_pot[ ss ] );
         fprintf( stdout, "\n" );
      }
   }
   ierr = MPI_Bcast( charges_pos, Ndim * N_charges, MPI_INT, 0, RGX_Comm );
   ierr = MPI_Bcast( charges_pot, N_charges, MD_type, 0, RGX_Comm );

#  if defined( DEBUG_stdout )
      my_tid_debug = my_tid
      fprintf( stdout, "Task %d's charges Info : \n", my_tid_debug );
      for ( ss = 0; ss < N_charges; ss++ )
      {
         for ( idir = 0; idir < Ndim; idir++ )
            fprintf( stdout, "%d ", charges_pos[ ss * Ndim + idir ] );
         fprintf( stdout, ", %f \n", charges_pot[ ss ] );
      }
#  endif
   

/* Decompose the global grid into local/task sub-grid and                  */
/* Determine the end-point of the local/task sub-grid, task_Grid_endpts[]  */

   ierr = RGX_Find_Task_Grid_endpts( RGX_Comm, Ndim, Glb_Grid_sz,
                                     task_Grid_endpts );

/* Allocate memory for Task_xGrid[] and Task_Grid_Modify[]             */

   hyper_vol = 1;
   for ( idir = 0; idir < Ndim; idir++ )
   {
      task_bndry_lgths[ idir ] = task_Grid_endpts[ idir + Ndim ]
                               - task_Grid_endpts[ idir ] + 1
                              + 2 * stride_width[ idir ];
      hyper_vol *= task_bndry_lgths[ idir ];
   }

   Task_xGrid        = ( double* ) calloc( hyper_vol, sizeof( double ) );
   Task_Grid_Modify  = ( int* )    calloc( hyper_vol, sizeof( int ) );

/* Define the "stride" for the ghost-point region                      */

   ierr = RGX_Defn_Task_Bndry_Strides( MD_type, Ndim, task_Grid_endpts,
                                       stride_width, task_bndry_strides );

/* Locate my nearest neighbor sub-grid's rank for exchange of Bndry_Strides */

   ierr = RGX_Find_Task_Grid_Ngbrs( RGX_Comm, Ndim, task_Ngbrs );

/* List all the processor allocated */

   ierr = RGX_List_Processor_Names( RGX_Comm );

/* Clean up the debugging file */

#  if defined ( DEBUG )
      my_tid_debug = my_tid;
      sprintf( filename_debug, filename_debug_fmt, my_tid_debug );
      sprintf( dbg_cmd_str, "rm -f " );
      strcat( dbg_cmd_str, filename_debug );
      ierr_debug = system( dbg_cmd_str );
#  endif

#  if defined ( TABLE )
      my_tid_table = my_tid;
      sprintf( filename_table, filename_table_fmt, my_tid_table );
      sprintf( table_cmd_str, "rm -f " );
      strcat( table_cmd_str, filename_table );
      ierr_table = system( table_cmd_str );
      file_table = fopen( filename_table, "a+" );
#  endif

#  if defined ( GPLOT )
      my_tid_gplot = my_tid;
      intvl_gplot = iter_Max / N_gplot;
#  endif

#  if defined ( DEMO )
      my_tid_demo = my_tid;
      intvl_demo = iter_Max / N_demo;
#  endif

#  if defined ( GPLOT )
      sprintf( dirname_gplot,  gplot_data_dir );
      sprintf( filename_gplot, filename_gplot_fmt, my_tid_gplot );

      /*  clear up the directroy for gplot-data  */
      if ( my_tid_gplot == 0 ) 
      {
         /* Exec "rm -f dirname_gplot"  */
         sprintf( gplot_cmd_str, "rm -rf " );
         strcat( gplot_cmd_str, dirname_gplot );
         ierr_gplot = system( gplot_cmd_str );

         /* Exec "mkdir dirname_gplot"  */
         sprintf( gplot_cmd_str, "mkdir " );
         strcat( gplot_cmd_str, dirname_gplot );
         ierr_gplot = system( gplot_cmd_str );
      }
#  endif

/* Initialize the local N-dim. grid  */

   ierr = MPI_Barrier( RGX_Comm );
   fbackground = 0.0;

   Init_Task_Grid_demo( Ndim, Glb_Grid_sz, task_Grid_endpts,
                        stride_width, &fbackground, MD_type,
                        N_charges, charges_pos, charges_pot,
                        Task_Grid_Modify, Task_xGrid );

#  if defined ( TABLE )
      fprintf( file_table, "Iteration %d : \n", 0 );
      print_Task_Grid( file_table, Ndim, task_bndry_lgths,
                       MD_type, Task_xGrid );
      fprintf( file_table, "\n" );
#  endif

#  if defined ( GPLOT )
   /* Output the results in a column form to be rendered by GNUplot  */
      iter_idx = 0;
      sprintf( sub_dirname_gplot, iter_fmt, iter_idx );
      if ( iter_idx % intvl_gplot == 0 )
         wrt_gplot_Task_Grid( my_tid_gplot,
                              dirname_gplot, sub_dirname_gplot,
                              filename_gplot, RGX_Comm,
                              Ndim, task_Grid_endpts,
                              stride_width, MD_type, Task_xGrid );
#  endif

#  if defined ( GPLOT ) && defined( DEMO )
      frame_offset[ 0 ] = 0; frame_offset[ 1 ] = 0;
      exe_gplot_Task_Grid( my_tid, iter_idx,
                           dirname_gplot, sub_dirname_gplot,
                           filename_gplot, Ndim, frame_offset );
#  endif

   for ( iter_idx = 1; iter_idx <= iter_Max; iter_idx++ )
   {

       ierr = RGX_Swap_Task_Bndry_Strides( RGX_Comm, MD_type,
                                           Task_xGrid, Ndim,
                                           task_Grid_endpts,
                                           stride_width,
                                           task_bndry_strides,
                                           task_Ngbrs );
       Laplace_Iter_incre( Ndim, Glb_Grid_sz, task_Grid_endpts,
                           stride_width, MD_type,
                           Task_Grid_Modify, Task_xGrid );

#     if defined ( GPLOT )
      sprintf( sub_dirname_gplot, iter_fmt, iter_idx );
      if ( iter_idx % intvl_gplot == 0 )
         wrt_gplot_Task_Grid( my_tid_gplot,
                              dirname_gplot, sub_dirname_gplot,
                              filename_gplot, RGX_Comm,
                              Ndim, task_Grid_endpts,
                              stride_width, MD_type, Task_xGrid );
#     endif

#     if defined ( TABLE )
         fprintf( file_table, "Iteration %d : \n", iter_idx );
         print_Task_Grid( file_table, Ndim, task_bndry_lgths,
                          MD_type, Task_xGrid );
         fprintf( file_table, "\n" );
#     endif

#     if defined ( DEMO )
         if (    my_tid_demo == 0 
              && iter_idx % intvl_demo == 0 
              && print_status_flag )
         {
            print_status_flag = 0;
            printf( "*** Current processes related to MPI "
                    "at iter_idx = %d : \n", iter_idx );
            printf( "*** EXEC \" vpm_cmd ps auxw \"|\" grep -i ASCI \"|\" "
                    "grep -v grep \" \n" );
            printf( "\n" );
            fflush( stdout );
            ierr_demo = system( "vpm_cmd ps auxw \"|\" grep -i mpi "
                                "\"|\" grep -v grep " );
            fflush( stdout );
            printf( "\n" );
            printf( "*** Current processes consumed more than "
                    "0:01 CPU time at iter_idx = %d : \n", iter_idx );
            printf( "*** EXEC \" vpm_cmd cpuhog -auxw 0:01 \" \n" );
            printf( "\n" );
            fflush( stdout );
            ierr_demo = system( "vpm_cmd cpuhog -auxw 0:01" );
            fflush( stdout );
            printf( "\n" );
            printf( "\n" );
         }
#     endif
       
   }   /*   for ( iter_idx = 0; iter_idx < 2; iter_idx++ )   */

#  if defined ( GPLOT ) && defined( DEMO )
      frame_offset[ 0 ] = 1024; frame_offset[ 1 ] = 0;
      exe_gplot_Task_Grid( my_tid, iter_idx,
                           dirname_gplot, sub_dirname_gplot,
                           filename_gplot, Ndim, frame_offset );
# endif

#  if defined( DEBUG_stdout )
      print_Task_Grid( stdout, Ndim, task_bndry_lgths, MD_type, Task_xGrid );
#  endif

   for ( idir = 0; idir < Ndim; idir++ )
      ierr = MPI_Type_free( &task_bndry_strides[ idir ] );

   ierr = MPI_Finalize();

#  if defined( TABLE )
      fclose( file_table );
#  endif
}  /*   END of the Program  */
